/****************************************************************************    
  Copyright (c) 1999,2000 WU-FTPD Development Group.  
  All rights reserved.
   
  Portions Copyright (c) 1980, 1985, 1988, 1989, 1990, 1991, 1993, 1994  
    The Regents of the University of California. 
  Portions Copyright (c) 1993, 1994 Washington University in Saint Louis.  
  Portions Copyright (c) 1996, 1998 Berkeley Software Design, Inc.  
  Portions Copyright (c) 1989 Massachusetts Institute of Technology.  
  Portions Copyright (c) 1998 Sendmail, Inc.  
  Portions Copyright (c) 1983, 1995, 1996, 1997 Eric P.  Allman.  
  Portions Copyright (c) 1997 by Stan Barber.  
  Portions Copyright (c) 1997 by Kent Landfield.  
  Portions Copyright (c) 1991, 1992, 1993, 1994, 1995, 1996, 1997  
    Free Software Foundation, Inc.    
   
  Use and distribution of this software and its source code are governed   
  by the terms and conditions of the WU-FTPD Software License ("LICENSE").  
   
  If you did not receive a copy of the license, it may be obtained online  
  at http://www.wu-ftpd.org/license.html.  
   
  $Id: config.fbs,v 1.16 2000/07/01 18:03:07 wuftpd Exp $  
   
****************************************************************************/
/* 
 * Configuration file for FreeBSD 2.X
 */
#define L_FORMAT "qd"
#define T_FORMAT "ld"
#define HAVE_LSTAT
#define HAVE_DIRENT_H
#undef  HAVE_D_NAMLEN
#define HAVE_FCNTL_H
#define HAVE_FLOCK
#undef  HAVE_FTW
#define	HAVE_GETCWD
#define HAVE_GETDTABLESIZE
#undef  HAVE_PSTAT
#define HAVE_REGEX_H
#define HAVE_ST_BLKSIZE
#undef  HAVE_SYSINFO
#define HAVE_SYSCONF
#define HAVE_UT_UT_HOST
#define USE_VAR
#define USE_LOG
#define VAR_RUN
#if !defined(USE_ETC_FTPD) && !defined(USE_ETC) && !defined(USE_OPT_FTPD)
#define USE_LOCAL_ETC
#endif
#define HAVE_SYS_MOUNT
#define HAVE_VPRINTF
#define HAVE_SNPRINTF
#define HAVE_REGEXEC
#undef  SHADOW_PASSWORD
#undef  USG
#define VIRTUAL
#if defined(__FreeBSD__)
#undef SPT_TYPE
#if __FreeBSD__ >= 2
#include <osreldate.h>		/* and this works */
#if __FreeBSD_version >= 199512	/* 2.2-current right now */
#define SPT_TYPE    SPT_BUILTIN
#endif
#endif
#ifndef SPT_TYPE
#define SPT_TYPE     SPT_REUSEARGV
#define SPT_PADCHAR  '\0'	/* pad process title with nulls */
#endif
#endif

#include <stdlib.h>
#include <unistd.h>

#ifndef FACILITY
#define FACILITY LOG_FTP
#endif

typedef void SIGNAL_TYPE;

#include "../config.h"
/* SKEY should always be defined on FreeBSD */
#ifndef SKEY
#define SKEY
#endif
#ifdef SKEY
#define SKEY_NAME
#endif
